/* GAME.PIKE - Started programming */
/* 0404: Started programming, added different textures and models for players,
         Made the room objects that are loaded, change some get_y() into
         get_effect(), added the "say" function, let the player describe
         themselves, adding init_game() call, added ghost movements, added item
         adding, checked the ghost through wall thing, added the who_list() */
/* 0504: Added the pick_up feature, added sizes to the inventory and who list,
         expanded the add_item() functionality, added the remove_living()
         function, added who to hit_player() */
/* 0704: Added shout, examine_closest(), room check */
/* 1004: Added size to monster_attack(), allowed multiple livings with the
         same original name, expanded attack to have who, check for no_fight */
/* 1104: Added different attack sounds, made say subscribable */
/* 1204: Expanded subscriptions */
/* 0205: Added spell command, spell sounds, show surround, preload rooms,
         monster spells, -40 level */
/* 040305: Added point light */
/* 070305: Added post office stuff */
/* 080305: Added top list stuff */
/* 160305: Added shutdown */
/* 260305: Find closest living */
/* 270305: Frogified "say", "tell" and "shout" */
/* 060405: Fixed torch issues for frog and removed preloads */

object *players,*items,*livings,pointlight,postoffice;
string toplist;
mapping rooms,zdir,xdir;
int unique;
mapping subscriptions;

void create() {
  call_out(renew_top_list,100);
  pointlight=(object)"pointlight.pike";
  pointlight->set_game(this_object());
  subscriptions=([]);
  players=({});
  livings=({});
  items=({});
  rooms=([]);
  zdir=(["north":1,"south":-1,"northwest":1,"northeast":1,"southeast":-1,
         "southwest":-1]);
  xdir=(["east":1,"west":-1,"northwest":-1,"northeast":1,"southeast":1,
         "southwest":-1]);
  unique=1;
  /* Preload rooms that have useful functionality */
  postoffice=load_room("rooms/village500499");
}

void point_light_change() {
int i;
string msg;

  for(i=0;i<sizeof(players);i++) {
    msg=pointlight->player_point_light(players[i],0);
    if(msg) players[i]->add_message(msg);
  }
}

object get_pointlight() {
  return pointlight;
}

void add_subscription(string f,object what) {
  if(!subscriptions[f]) {
    subscriptions[f]=({what});
    return;
  }
  /* Avoid duplicates */
  subscriptions[f]-=({what});
  subscriptions[f]+=({what});
  /* Remove dead objects */
  subscriptions[f]-=({0});
}

int distance(object a,object b) {
int x,y,z;

  if(a->get_room()!=b->get_room()) return 10000002;
  x=(a->get_x())-(b->get_x());
  x=x*x;
  y=(a->get_y())-(b->get_y());
  y=y*y;
  z=(a->get_z())-(b->get_z());
  z=z*z;
  return x+y+z;
}

string examine_closest(object player) {
object closest,room;
int dist,maxdist,i;
mixed *roomclose;

  maxdist=1000001;
  for(i=0;i<sizeof(livings);i++) {
    dist=distance(livings[i],player);
    if(dist<maxdist) {
      maxdist=dist;
      closest=livings[i];
    }
  }
  for(i=0;i<sizeof(players);i++) {
    if(players[i]==player) continue;
    dist=distance(players[i],player);
    if(dist<maxdist) {
      maxdist=dist;
      closest=players[i];
    }
  }
  for(i=0;i<sizeof(items);i++) {
    dist=distance(items[i],player);
    if(dist<maxdist) {
      maxdist=dist;
      closest=items[i];
    }
  }
  /* The room might have something closer */
  roomclose=(player->get_room())->closest_view(player);
  if(roomclose[0]<maxdist) return roomclose[1];

  /* Otherwise, display what was found */
  if(maxdist>1000000) return "008 Nothing here.\n";
  return "004 CAMERA EYE "+player->get_x()+","+(player->get_y()+10)+","
       + player->get_z()+" AT "+closest->get_x()+","+closest->get_y()+","
       + closest->get_z()+"\n"+closest->get_description();
}

void remove_living(object who) {
  livings-=({who});
  destruct(who);
}

string who_list() {
string result;
object current;
int i;

  result="016 SIZE "+sizeof(players)+"\n";
  for(i=0;i<sizeof(players);i++) {
    result+="016 WHO "+players[i]->get_name();
    if(players[i]->get_idle()) result+=" (idle)";
    result+="\n";
  }
  return result;
}

void add_living(object who) {
string name;

  name=who->get_name();
  who->set_short(name);
  name=""+unique+"#"+name;
  who->set_name(name);
  unique++;
  livings+=({who});
  say(who,who->get_monster_description());
}

void add_item(object what) {
string name,short;
int i,found;

  /* First check if the item is already in our list */
  found=-1;
  for(i=0;i<sizeof(items);i++) {
    if(items[i]==what) found=i;
  }
  if(found==-1) {
    /* This is a new item, make the item name unique */
    name=what->get_name();
    short=what->get_short();
    what->set_short(short);
    name=""+unique+"#"+name;
    what->set_name(name);
    unique++;
    /* Add to list */
    items+=({what});
  }
  /* Indicate its presence to the world */
  say(what,
      "001 NAME "+what->get_name()+" MODEL "+what->get_model()+" TEXTURE "
    + what->get_texture()+"\n003 LOCATE "+what->get_name()+" X "+what->get_x()
    + " Y "+what->get_y()+" Z "+what->get_z()+"\n");
}

void destruct_object(object what) {
  destruct(what);
}

void remove_item(object what) {
  say(what,"015 REMOVE "+what->get_name()+"\n");
  items-=({what});
  /* Don't destruct immediately, as it might have requested it itself */
  call_out(destruct_object,0,what);
}

object *livings_in_area(int x1,int z1,int x2,int z2) {
int i,x,z;
object *ps;

  ps=({});
  for(i=0;i<sizeof(players);i++) {
    x=players[i]->get_x();
    z=players[i]->get_z();
    if((x>=x1)&(x<=x2)&(z>=z1)&(z<=z2)) ps+=({players[i]});
  }
  for(i=0;i<sizeof(livings);i++) {
    x=livings[i]->get_x();
    z=livings[i]->get_z();
    if((x>=x1)&(x<=x2)&(z>=z1)&(z<=z2)) ps+=({livings[i]});
  }
  if(sizeof(ps)==0) return 0;
  return ps;
}

object living_present(int x,int z) {
int i;

  for(i=0;i<sizeof(players);i++) {
    if(players[i]->occupies(x,z)) return players[i];
  }
  for(i=0;i<sizeof(livings);i++) {
    if(livings[i]->occupies(x,z)) return livings[i];
  }
  return 0;
}

void end_spell(object who,object victim,object which) {
string result;

  result="015 REMOVE "+which->get_name()+"\n";
  if(victim) result+="010 pettchew\n";
  say(who,result);
  if(victim) {
    victim->hit_player(random(6)+1,who);
  }
  pointlight->remove_point_light_object(which);
}

int monster_spell(object who,object victim) {
object room,spell;
int x,y,z,xinit,yinit,zinit,dx,dy,dz;
string result;

  room=who->get_room();
  if(room) if(room->get_property("no_fight")) return 0;
  /* Ghosts can't be attacked */
  if(victim->is_ghost()) return 0;
  x=victim->get_x();
  y=victim->get_y();
  z=victim->get_z();
  xinit=who->get_x();
  yinit=who->get_y();
  zinit=who->get_z();
  dx=x-xinit;
  dz=z-zinit;
  if(dx*dx+dz*dz>225) return 0;
  /* Target location is x,z, victim is victim */
  result="010 down\n";
  result+="001 NAME spell"+unique+" MODEL spell.txt TEXTURE items8_.tm2\n";
  result+="003 LOCATE spell"+unique+" X "+xinit+" Y "+yinit+" Z "+zinit+"\n";
  result+="017 ROTATEY spell"+unique+"\n";
  result+="022 SPEED spell"+unique+" 3\n";
  result+="003 LOCATE spell"+unique+" X "+x+" Y "+y+" Z "+z+"\n";
  spell=clone((program)"spot");
  spell->set_name("spell"+unique);
  spell->set_xyz(x,y,z);
  pointlight->add_point_light_object(spell);
  call_out(end_spell,1,who,victim,spell);
  unique++;
  say(who,result);
  return 1;
}
 
string spell(object who,string where) {
int x,y,z,dist,xinit,zinit,yinit,ynew,i;
object victim,room,spell;
string result;

  if(who->get_attacking()) return "995 Busy\n";
  /* You don't get a second chance... Even if it was a bad attack attempt */
  who->set_attacking();
  room=who->get_room();
  if(room) if(room->get_property("no_fight")) {
    return "008 You don't want to fight here.\n";
  }
  if(who->get_level()<5) {
    return "008 Your level is too low for magic missile.\n";
  }
  if((who->cast_spell(10))<0) return "008 Your magic missile spell failed.\n";
  if(who->is_ghost()) return "992 Are ghost\n";
  x=who->get_x();
  z=who->get_z();
  y=who->get_y()+1;
  for(i=0;i<15;i++) {
    x=x+xdir[where];
    z=z+zdir[where];
    if(!i) {
      xinit=x;
      zinit=z;
    }
    victim=living_present(x,z);
    if(victim) {
      y=victim->get_y()+2;
      break;
    }
    ynew=room->get_y(x,z)+1;
    if((ynew<0)&&(ynew>-20)) break;
    y=ynew;
    if(!i) {
      yinit=y;
    }
  }
  /* Target location is x,z, victim is victim */
  result="010 down\n";
  result+="001 NAME spell"+unique+" MODEL spell.txt TEXTURE items8_.tm2\n";
  result+="003 LOCATE spell"+unique+" X "+xinit+" Y "+yinit+" Z "+zinit+"\n";
  result+="017 ROTATEY spell"+unique+"\n";
  result+="022 SPEED spell"+unique+" 3\n";
  result+="003 LOCATE spell"+unique+" X "+x+" Y "+y+" Z "+z+"\n";
  spell=clone((program)"spot");
  spell->set_name("spell"+unique);
  spell->set_xyz(x,y,z);
  pointlight->add_point_light_object(spell);
  call_out(end_spell,1,who,victim,spell);
  unique++;
  say(who,result);
  return "008 You succesfully cast magic missile.\n";
}

/* 
 * NOTE: This function returns negative values if the spot is not clean, and 0
 *       or positive values in case it is clean. The return value is the proper
 *       y.
 */
int spot_clean(object who,int x,int z) {
int i,result;
object room;

  if(living_present(x,z)) return -1;
  room=who->get_room();
  /* Assume it is occupied if there is no room */
  if(!room) return -1;
  for(i=0;i<sizeof(items);i++) {
    if(items[i]->get_room()!=room) continue;
    if(items[i]->occupies(x,z)) return -1;
  }
  result=room->get_real_effect(x,z);
  if(!result) return 0;
  if(result=='0') return 0;
  if(result<'a') return -1;
  return result-'a';
}

object drop_spot(object who) {
int x,y,z;
object spot;

  x=who->get_x()-1;
  z=who->get_z()-1;
  y=spot_clean(who,x,z);
  if(y<0) {
    z++;
    y=spot_clean(who,x,z);
  }
  if(y<0) {
    z++;
    y=spot_clean(who,x,z);
  }
  if(y<0) {
    x++;
    y=spot_clean(who,x,z);
  }
  if(y<0) {
    z-=2;
    y=spot_clean(who,x,z);
  }
  if(y<0) {
    x++;
    y=spot_clean(who,x,z);
  }
  if(y<0) {
    z++;
    y=spot_clean(who,x,z);
  }
  if(y<0) {
    z++;
    y=spot_clean(who,x,z);
  }
  if(y<0) return 0;
  spot=clone((program)"spot");
  if(y==25) y=-40;
  spot->set_xyz(x,y,z);
  return spot;
}

int spot_occupied(object who,int x,int z) {
int i,result;
object room,living;

  living=living_present(x,z);
  if((living)&&(living!=who)) return 1;
  room=get_room(who,x,z);
  /* Assume it is occupied if there is no room */
  if(!room) {
    werror("No room could be loaded, therefore access as denied.\n");
    return 1;
  }
  result=room->get_effect(x,z);
  /* If it is -1 then the spot is blocked by something */
  if(result==-1) return 1;
  /* Special ghost value */
  if(result==-4) {
    if(who->is_ghost()) return 2;
    return 1;
  }
  /* If this isn't a player, no further checks are needed */
  if(!who->is_player()) return 0;
  /* If it is -2 then the room needs to talk to the player */
  if(result==-2) {
    room->trigger_hit(who);
  }
  else {
    /* If it isn't -2, then there obviously is no trigger */
    who->remove_room_trigger();
  }
  return 0;
}

void find_free_spot(object who) {
int x,y,z,i,found;
object room;

  x=150;
  z=150;
  found=0;
  who->set_x(x);
  who->set_z(z);
  who->set_y(0);
  who->set_location("village");
  while(!found) {
    if(spot_occupied(who,x,z)) {
      if(x>160) {
        x=150;
        z++;
      }
      else {
        x++;
      }
    }
    else {
      found=1;
    }
  }
  who->set_x(x);
  who->set_z(z);
  room=get_room(who,x,z);
  if(room) {
    y=room->get_y(x,z);
    who->set_y(y);
  }
  else {
    who->set_y(0);
  }
}

void shout(string message) {
int i;

  for(i=0;i<sizeof(players);i++) players[i]->add_message(message);
}

void say(object who,string message,void|object exclude) {
int i;

  for(i=0;i<sizeof(players);i++) {
    if(players[i]==exclude) continue;
    if(players[i]->get_room()!=who->get_room()) continue;
    players[i]->add_message(message);
  }
}

string tell_player(object who,string target,string message) {
int i;
object to;

  if((who->cast_spell(4))<0) {
    return "008 Your tell spell failed.\n";
  }
  for(i=0;i<sizeof(players);i++) {
    if(players[i]->get_name()==target) to=players[i];
  }
  if(!to) return "994 No such player\n";
  to->add_message("008 "+who->get_name()+" tells you: "+message+"\n");
  return "008 You tell "+target+": "+message+"\n";
}

void exit_now() {
  exit(1);
}

void shut_down() {
int i;

  for(i=0;i<sizeof(players);i++) players[i]->save();
  shout("008 Shutdown now... Saving player characters.\n");
  call_out(exit_now,1);
}

string debug_action(object who,string what) {
string *words;
int x,z;

  words=what/" ";
  if(words[0]=="goto") {
    x=(int)words[1];
    z=(int)words[2];
    get_room(who,x,z);
    who->set_x(x);
    who->set_z(z);
    return "008 So moved.\n";
  }
  if(words[0]=="shutdown") {
    shout("008 Shutdown initiated. Game will reboot in 60 seconds.\n"
        + "008 DISCONNECT NOW!\n");
    call_out(shut_down,60);
    return "008 Ok.\n";
  }
  return "008 Debug disabled!\n";
}

string shout_text(object who,string message) {
  if(who->cast_spell(50)<0) {
    return "008 Your shout spell failed.\n";
  }
  if(who->is_frog()) {
    shout("008 "+who->get_short()+" shouts: Hribbit!\n010 frog\n");
    return "008 You make a loud frog noise.\n";
  }
  shout("008 "+who->get_short()+" shouts: "+message+"\n010 shout\n");
  return "008 You shouted.\n";
}

string say_text(object who,string message) {
object target,*listeners;
int i;
string result;

//  if(message[0..0]=="/") return debug_action(who,message[1..]);
  listeners=subscriptions["say_text"];
  if(listeners) {
    for(i=0;i<sizeof(listeners);i++) {
      target=listeners[i];
      result=target->say_text(who,message);
      if(result) return result;
    }
  }
  if(who->is_frog()) message="Hribbit!\n010 frog";
  say(who,"008 "+who->get_name()+" says: "+message+"\n",who);
  return "008 You say: "+message+"\n";
}

object add_player(object who,string name) {
int i;
object np;

  for(i=0;i<sizeof(players);i++) {
    if(players[i]->get_name()==name) {
      players[i]->set_connection(who);
      return players[i];
    }
  }
  np=clone((program)"player");
  np->load("/home/mhofman/Pike/Server/players/"+name+".o");
  np->set_name(name);
  np->set_connection(who);
  players+=({np});
  find_free_spot(np);
  say(np,np->get_player_description(),np);
  return np;
}

/* 1: players, 2: livings, 3: both */
object find_closest(object from,int what) {
int newdist,distance,i;
object found;

  distance=10000;
  found=0;
  if(what&1) {
    for(i=0;i<sizeof(players);i++) {
      if(players[i]==from) continue;
      newdist=players[i]->get_x()-from->get_x();
      newdist*=newdist;
      newdist+=(players[i]->get_z()-from->get_z())
             * (players[i]->get_z()-from->get_z());
      if(newdist<distance) {
        found=players[i];
        distance=newdist;
      }
    }
  }
  if(what&2) {
    for(i=0;i<sizeof(livings);i++) {
      if(livings[i]==from) continue;
      newdist=livings[i]->get_x()-from->get_x();
      newdist*=newdist;
      newdist+=(livings[i]->get_z()-from->get_z())
             * (livings[i]->get_z()-from->get_z());
      if(newdist<distance) {
        found=livings[i];
        distance=newdist;
      }
    }
  }
  return found;
}

object load_room(string fn) {
object room;

  werror("Loading room "+fn+"!\n");
  room=(object)(fn+".pike");
  room->set_game(this_object());
  if(!room) return 0;
  werror("Room object succesfully created.\n");
  if(!room->load_file(fn)) return 0;
  werror("Room succesfully loaded.\n");
  rooms[fn]=room;
  return room;
}

object get_room(object who,int x,int z) {
object room;
string fn;
int cx,cz;

  cx=x/100+500;
  cz=z/100+500;
  fn="rooms/"+who->get_location()+cx+""+cz;
  if(rooms[fn]) {
    who->set_room(rooms[fn]);
    return rooms[fn];
  }
  room=load_room(fn);
  who->set_room(room);
  return room;
}

void recheck_light(object room) {
object cr;
int i;

  for(i=0;i<sizeof(players);i++) {
    cr=players[i]->get_room();
    if(!cr) continue;
    if(cr!=room) continue;
    if(cr->get_light(players[i])) {
      players[i]->add_message("024 LIGHT 1\n");
      continue;
    }
    players[i]->add_message("024 LIGHT 0\n");
  }
}

int torch_check(object room) {
int i;

  for(i=0;i<sizeof(items);i++) {
    if(items[i]->get_type()=="torch") {
      if(room->torch_check(items[i])) return 1;
    }
  }
  return 0;
}

string look(object who) {
string result,roomname,roombase,msg;
int i,rx,rz,j;
object room,surround;

  result="002 Description start\n";
  room=who->get_room();
  if(!room) werror("No room?\n");
  if(!room->get_light(who)) result+="024 LIGHT 0\n";
  result+=room->get_description();
  /* Also add the descriptions of surrounding rooms, if available */
  roomname=room->get_room_name();
  roombase="rooms/"+who->get_location();
  roomname-=roombase;
  rx=(int)roomname[0..2];
  rz=(int)roomname[3..];
  for(i=rx-1;i<rx+2;i++) {
    for(j=rz-1;j<rz+2;j++) {
      if((i==rx)&&(j==rz)) continue;
      surround=rooms[roombase+i+j];
      if(!surround) continue;
      if(who->room_shown(surround)) continue;
      result+=surround->get_description();
    }
  }
  /* End of also adding the descriptions of surrounding rooms */
  for(i=0;i<sizeof(players);i++) {
    /* Only add players that are in the same room */
    if(players[i]->get_room()==room) {
      result+=players[i]->get_player_description();
    }
  }
  for(i=0;i<sizeof(livings);i++) {
    /* Only add livings that are in the same room */
    if(livings[i]->get_room()==room) {
      result+=livings[i]->get_monster_description();
    }
  }
  for(i=0;i<sizeof(items);i++) {
    /* Only add items that are in the same room */
    if(items[i]->get_room()==room) {
      result+="001 NAME "+items[i]->get_name();
      result+=" MODEL "+items[i]->get_model()+" TEXTURE ";
      result+=items[i]->get_texture()+"\n";
      result+="003 LOCATE "+items[i]->get_name()+" X "+items[i]->get_x();
      result+=" Y "+items[i]->get_y()+" Z "+items[i]->get_z()+"\n";
    }
  }
  result+=who->describe();
  /* Check if there is mail */
  if(postoffice->check_mail(who)) {
    result+="008 You have NEW mail at the post office.\n010 newmail\n";
  }
  /* Add point lights, if available */
  msg=pointlight->player_point_light(who,1);
  if(msg) result+=msg;
  result+="005 Description end\n";
  return result;
}

int monster_attack(object who,object victim,int size) {
int dx,dz,damage;
object room;
  
  room=who->get_room();
  if(room) if(room->get_property("no_fight")) return 0;
  /* Ghosts can't be attacked */
  if(victim->is_ghost()) return 0;
  dx=who->get_x()-victim->get_x();
  if(dx<-size) return 0;
  if(dx>size) return 0;
  dz=who->get_z()-victim->get_z();
  if(dz<-size) return 0;
  if(dz>size) return 0;
  damage=who->attack(victim);
  say(who,"013 ANIMATE "+who->get_name()+"2\n");
  if(!damage) {
    return 1;
  }
  /* Monsters do not get a master strike */
  if(damage<0) damage=-damage;
  if(victim->defend()) {
    return 2;
  }
  damage-=victim->get_armour_class();
  if(damage<=0) {
    return 3;
  }
  victim->add_message("008 "+who->get_short()+" hits you.\n");
  victim->hit_player(damage,who);
  return 4;
}

string attack(object who,string where) {
int x,z,damage,victim_level,isplayer;
object victim,room;
string name,attacksound;

  if(who->get_attacking()) return "995 Busy\n";
  /* You don't get a second chance... Even if it was a bad attack attempt */
  who->set_attacking();
  room=who->get_room();
  if(room) if(room->get_property("no_fight")) {
    return "008 You don't want to fight here.\n";
  }
  if(who->is_ghost()) return "992 Are ghost\n";
  x=who->get_x();
  z=who->get_z();
  z+=zdir[where];
  x+=xdir[where];
  victim=living_present(x,z);
  if(victim==who) {
    werror("Attacking myself?\n");
    return "994 Server error\n";
  }
  if(!victim) {
    say(who,"010 miss\n013 ANIMATE "+who->get_name()+"3\n");
    return "993 No victim\n";
  }
  if(victim->is_ghost()) {
    say(who,"010 miss\n013 ANIMATE "+who->get_name()+"3\n");
    return "992 Is ghost\n";
  }
  say(who,"013 ANIMATE "+who->get_name()+"2\n");
  damage=who->attack(victim);
  if(!damage) return "010 miss\n990 Attack failed\n";
  if(damage>0) {
    if(victim->defend()) return "989 Attack fended off\n";
    damage-=victim->get_armour_class();
    if(damage<=0) return "988 Attack caught by armour\n";
  }
  else {
    /* Master attack, defense and AC don't count */
    damage=-damage;
  }
  /* Get name and level now, as the victim may die */
  name=victim->get_short();
  victim_level=victim->get_level();
  victim->hit_player(damage,who);
  isplayer=victim->is_player();
  attacksound=who->get_attack_sound();
  if(!attacksound) attacksound="punch";
  say(who,"010 "+attacksound+"\n");
  if((!victim)||(victim->is_ghost())) {
    if(isplayer) {
      shout("008 "+name+" died at the hands of "+who->get_short()+".\n");
    }
    else {
      say(who,"008 "+name+" died.\n");
    }
    who->add_experience((victim_level+1)*victim_level);
  }
  return "008 You hit "+name+"\n";
}

int monster_move(object who,int tx,int tz) {
int x,y,z,i;
object room;

  x=who->get_x();
  y=who->get_y();
  z=who->get_z();
  if(tx>x) x++;
  if(tx<x) x--;
  if(tz>z) z++;
  if(tz<z) z--;
  if(spot_occupied(who,x,z)) return 0;
  who->set_x(x);
  who->set_z(z);
  room=get_room(who,x,z);
  if(!room) {
    y=0;
    werror("No room found.\n");
  }
  else {
    y=room->get_y(x,z);
  }
  who->set_y(y);
  for(i=0;i<sizeof(players);i++) {
    if(players[i]==who) continue;
    players[i]->add_message("003 LOCATE "+who->get_name()+" X "+x+" Y "+y
                          + " Z "+z+"\n");
  }
  return 1;
}

string move(object who,string where) {
int x,y,z,i;
object room;

  x=who->get_x();
  y=who->get_y();
  z=who->get_z();
  z+=zdir[where];
  x+=xdir[where];
  /* Float the ghost */
  if(who->is_ghost()) {
    while(spot_occupied(who,x,z)==2) {
      z+=zdir[where];
      x+=xdir[where];
    }
  }
  if(spot_occupied(who,x,z)) {
    return "997 Spot occupied\n";
  }
  who->set_x(x);
  who->set_z(z);
  room=get_room(who,x,z);
  if(!room) {
    y=0;
    werror("No room found.\n");
  }
  else {
    y=room->get_y(x,z);
  }
  who->set_y(y);
  for(i=0;i<sizeof(players);i++) {
    if(players[i]==who) continue;
    players[i]->add_message("003 LOCATE "+who->get_name()+" X "+x+" Y "+y
                          + " Z "+z+"\n");
  }
  who->remove_item_trigger();
  for(i=0;i<sizeof(items);i++) {
    if(!items[i]->can_pick_up()) continue;
    if(items[i]->get_room()!=room) continue;
    if((items[i]->get_x()==x)&&(items[i]->get_z()==z)) {
      who->add_item_trigger(items[i]);
      break;
    }
  }
  return "003 LOCATE "+who->get_name()+" X "+x+" Y "+y+" Z "+z+"\n";
}

string help(string what) {
string data;

  data=Stdio.read_file("help/"+what);
  if(!data) return "994 Not found\n";
  return data;
}


void renew_top_list() {
  top_list();
  call_out(renew_top_list,4000);
}

string get_top_list() {
  if(!toplist) return top_list();
  return toplist;
}

string top_list() {
string *directory,data,*names,name,*lines;
int *xps,*levels,xp,level,i,j;

  xps=({});
  levels=({});
  names=({});
  directory=get_dir("players");
  for(i=0;i<sizeof(directory);i++) {
    data=Stdio.read_file("players/"+directory[i]);
    if(data) {
      lines=data/"\n";
      xp=0;
      level=1;
      name=0;
      for(j=0;j<sizeof(lines);j++) {
        sscanf(lines[j],"name=%s",name);
        sscanf(lines[j],"xp=%d",xp);
        sscanf(lines[j],"level=%d",level);
      }
      if(name) {
        names+=({name});
        xps+=({xp});
        levels+=({level});
      }
    }
  }
  sort(xps,names,levels);
  toplist="";
  for(i=sizeof(names)-1;i>sizeof(names)-10;i--) {
    toplist+="016 TXT "+(sizeof(names)-i)+". "+names[i]+" ("+levels[i]+"): ";
    toplist+=xps[i]+"\n";
  }
  return toplist;
}

