/* PLAYER.PIKE - Programmed by Maarten Hofman */
/* 0404: Adding triggers, say things, hit points, resurrect, fighting */
/* 0504: Added object triggers, drop item, extracted living, changed texture
         and model, added who to hit_player(), gold coins */
/* 240704: Added update for knife arm */
/* 010804: Added player to use command in item */
/* 230804: Added handle_shop */
/* 140904: Added autosave functionality */
/* 190904: Added experience points to the score command */
/* 200904: Added advance commands */
/* 111004: Debugged the stuck armour */
/* 130105: Added present for specific item types */
/* 160205: Added room_shown functionality for look expansion */
/* 040305: No selling of lit torches! */
/* 060305: Added special item handling */
/* 090305: Added link dead gold stuff */
/* 160305: Added idle indicator */
/* 260305: Added frog stuff */
/* 270305: Make sure the frog is saved */
/* 060405: Fixed torch issues for frog and preload */
/* 080405: Made login/logout messages */

#include <simulate.h>

inherit "living";

/* Variables for the player part */
string trigname,password,oldtexture,oldmodel,email;
int ghost,xp,skills,quests,idle,frog;
object connection,roomtrigger,itemtrigger,*roomtrack,specitem;

void set_idle() {
  idle=1;
}

void reset_idle() {
  remove_call_out(set_idle);
  call_out(set_idle,300);
  idle=0;
}

int get_idle() {
  return idle;
}

int room_shown(object surround) {
int i;

  for(i=0;i<sizeof(roomtrack);i++) {
    if(surround==roomtrack[i]) return 1;
  }
  return 0;
}

int advance_score(string which) {
int needed,i;

  switch(lower_case(which)) {
    case "level":
    case "advance level":
      needed=0;
      for(i=0;i<level;i++) needed+=(i+1)*100;
      if(xp>=needed) {
        level++;
        needed=random(6)+1;
        max_hp+=needed;
        current_hp+=needed;
        needed=random(6)+1;
        max_sp+=needed;
        current_sp+=needed;
        skills+=1;
        return 1;
      }
      return 0;
    case "attack":
    case "advance attack":
      if(skills>0) {
        att++;
        skills--;
        return 1;
      }
      return 0;
    case "defense":
    case "advance defense":
      if(skills>0) {
        def++;
        skills--;
        return 1;
      }
  }
  return 0;
}

void auto_save() {
  if(current_hp<max_hp) current_hp++;
  if(current_sp<max_sp) current_sp++;
  save();
  call_out(auto_save,3600);
}

void create() {
  ::create();
  roomtrack=({});
  player=1;
  specitem=0;
  skills=0;
  texture="axes_.tm2";
  model="player.ani";
  call_out(auto_save,3600);
  call_out(set_idle,300);
}

void save() {
string fn,result;
int ag,i;

  /* 
   * Calculate additional gold in case the server goes down with items in your
   * hands.
   */
  ag=0;
  for(i=0;i<sizeof(items);i++) ag=ag+items[i]->get_value();
  fn="/home/mhofman/Pike/Server/players/"+name+".o";
  rm(fn);
  result="name="+name+"\n";
  result+="gender="+gender+"\n";
  result+="password="+password+"\n";
  result+="texture="+texture+"\n";
  result+="model="+model+"\n";
  result+="oldtexture="+oldtexture+"\n";
  result+="oldmodel="+oldmodel+"\n";
  result+="hp="+current_hp+"\n";
  result+="max_hp="+max_hp+"\n";
  result+="sp="+current_sp+"\n";
  result+="max_sp="+max_sp+"\n";
  result+="ghost="+ghost+"\n";
  result+="frog="+frog+"\n";
  result+="email="+email+"\n";
  result+="attack="+att+"\n";
  result+="defence="+def+"\n";
  result+="level="+level+"\n";
  result+="gold="+(gold+ag/4)+"\n";
  result+="xp="+xp+"\n";
  result+="skills="+skills+"\n";
  result+="quests="+quests+"\n";
  Stdio.write_file(fn,result);
}

void set_quest(int which) {
  werror("set_quest("+which+") called in "+name+"\n");
  quests|=which;
}

int get_quest(int which) {
  return quests&which;
}

void load(string fn) {
string data,*lines,*words;
int i,resave;

  data=Stdio.read_file(fn);
  if(!data) return;
  lines=data/"\n";
  resave=0;
  for(i=0;i<sizeof(lines)-1;i++) {
    words=lines[i]/"=";
    switch(words[0]) {
      case "name":
        set_name(words[1]);
        break;
      case "gender":
        if(words[1]=="female") {
          gender=2;
          texture="woman8_.tm2";
          model="blwf.ani";
          resave=1;
          break;
        }
        if(words[1]=="male") {
          gender=1;
          texture="woman8_.tm2";
          model="brwm.ani";
          resave=1;
          break;
        }
        gender=(int)words[1];
        break;
      case "model": model=words[1]; break;
      case "texture": texture=words[1]; break;
      case "hp": current_hp=(int)words[1]; break;
      case "max_hp": max_hp=(int)words[1]; break;
      case "sp": current_sp=(int)words[1]; break;
      case "max_sp": max_sp=(int)words[1]; break;
      case "ghost": ghost=(int)words[1]; break;
      case "frog": frog=(int)words[1]; break;
      case "attack": att=(int)words[1]; break;
      case "defence": def=(int)words[1]; break;
      case "level": level=(int)words[1]; break;
      case "gold": gold=(int)words[1]; break;
      case "xp": xp=(int)words[1]; break;
      case "skills": skills=(int)words[1]; break;
      case "oldtexture": oldtexture=words[1]; break;
      case "oldmodel": oldmodel=words[1]; break;
      case "quests": quests=(int)words[1]; break;
      case "email": email=words[1]; break;
      case "password":
        password=words[1];
        break;
      default:
        werror("Unknown line in save file: "+lines[i]+"\n");
        break;
    }
  }
  if(resave) save();
}

object present(string what) {
int i;
string *words;

  for(i=0;i<sizeof(items);i++) {
    words=(items[i]->get_short())/" ";
    if(words[0]==what) return items[i];
  }
}

object *type_present(string what) {
object *result;
int i;

  result=({});
  for(i=0;i<sizeof(items);i++) {
    if(items[i]->get_type()==what) result+=({items[i]});
  }
  if(sizeof(result)==0) return 0;
  return result;
}

void remove_item_from_list(object item) {
  items-=({item});
  if(weapon==item) {
    weapon=this_object();
    game->shout("014 UPDATE "+name+"2 MODEL warm.txt TEXTURE "+texture+"\n");
    item->use(0,this_object());
  }
  if(specitem==item) {
    specitem=0;
    game->shout("014 UPDATE "+name+"4 MODEL whand.txt TEXTURE "+texture+"\n");
  }
  if(armour==item) {
    if(gender==1) {
      game->shout("014 UPDATE "+name+"3 MODEL wmbody.txt TEXTURE "+texture+"\n");
    }
    else {
      game->shout("014 UPDATE "+name+"3 MODEL wfbody.txt TEXTURE "+texture+"\n");
    }
    armour=this_object();
    item->use(0,this_object());
  }
}

string drop_item(string what) {
object item,spot;
int i;

  werror("drop_item("+what+") called in player "+name+"\n");
  item=present(what);
  if(!item) return "994 Server error\n";
  if(item->cant_drop()==1) return "008 You don't want to drop that item.\n";
  if(item->cant_drop()==2) {
    remove_item_from_list(item);
    return "008 You dropped the "+what+" and it vanishes in thin air.\n";
  }
  spot=game->drop_spot(this_object());
  if(!spot) return "008 There is no room here.\n";
  item->set_x(spot->get_x());
  item->set_y(spot->get_y());
  item->set_z(spot->get_z());
  item->set_room(room);
  game->add_item(item);
  remove_item_from_list(item);
  return "008 You dropped the "+what+"\n";
}

void add_experience(int what) {
  xp+=what;
}

string use_item(string what) {
object item;
int i;

  werror("use_item("+what+") called in player "+name+"\n");
  item=0;
  for(i=0;i<sizeof(items);i++) {
    if((items[i]->get_short()/" ")[0]==what) item=items[i];
  }
  if(!item) return "994 Server error\n";
  if(!item->can_use()) {
    return "008 You can't use the "+what+"\n";
  }
  if(weapon==item) {
    weapon=this_object();
    item->use(0,this_object());
    game->shout("014 UPDATE "+name+"2 MODEL warm.txt TEXTURE "+texture+"\n");
    return "008 You unwield the "+what+"\n";
  }
  if(armour==item) {
    armour=this_object();
    item->use(0,this_object());
    if(gender==1) {
      game->shout("014 UPDATE "+name+"3 MODEL wmbody.txt TEXTURE "+texture+"\n");
    }
    else {
      game->shout("014 UPDATE "+name+"3 MODEL wfbody.txt TEXTURE "+texture+"\n");
    }
    return "008 You take off the "+what+"\n";
  }
  if(item->get_type()=="weapon") {
    weapon->use(0,this_object());
    weapon=item;
    item->use(1,this_object());
    game->shout("014 UPDATE "+name+"2 MODEL "+item->get_use_model()+" TEXTURE "
              + item->get_use_texture()+"\n");
    return "008 You wield the "+what+"\n";
  }
  if(item->get_type()=="armour") {
    armour->use(0,this_object());
    armour=item;
    item->use(1,this_object());
    game->shout("014 UPDATE "+name+"3 MODEL "+item->get_use_model(gender)
              + " TEXTURE "+item->get_use_texture()+"\n");
    return "008 You wear the "+what+"\n";
  }
  return item->use(this_object());
}

void set_room(object how) {
int i,same;

  same=0;
  if(room==how) same=1;
  ::set_room(how);
  for(i=0;i<sizeof(roomtrack);i++) {
    if(roomtrack[i]==how) {
      if(!same) add_message(how->get_update());
      return;
    }
  }
  add_message(game->look(this_object()));
  roomtrack+=({how});
}

int is_ghost() {
  return ghost;
}

int is_frog() {
  return frog;
}

string describe() {
  return "012 HP "+current_hp+" MAX "+max_hp+"\n";
}

void remove_room_trigger() {
  if(!roomtrigger) return;
  add_message("006 Room trigger removed\n");
  roomtrigger=0;
}

void add_room_trigger(object room,string trigger) {
  if(ghost) {
    /* Ghosts can't use triggers unless specified */
    if(lower_case(trigger[0..4])!="ghost") return;
  }
  add_message("007 "+trigger+"\n");
  trigname=trigger;
  roomtrigger=room;
}

void add_item_trigger(object item) {
  /* Ghosts can't pick up things */
  if(ghost) return;
  /* Frogs can't pick up things either */
  if(item->get_type()=="torch") if(frog) return;
  add_message("007 pick up "+item->get_short()+"\n");
  trigname="item";
  itemtrigger=item;
}

void remove_item_trigger() {
  if(!itemtrigger) return;
  add_message("006 Item trigger removed\n");
  itemtrigger=0;
}

string trigger_activated() {
  if(itemtrigger) {
    return itemtrigger->pick_up(this_object());
  }
  if(!roomtrigger) return "996 No trigger\n";
  return roomtrigger->trigger_activated(this_object(),trigname);
}

void add_message(string how) {
  if(connection) connection->add_message(how);
}

int cast_spell(int how) {
  /* Ghosts can't cast spells */
  if(ghost) return -1;
  if(how>current_sp) return -2;
  current_sp=current_sp-how;
  /* TODO: Add tests whether spell is succesful */
  return 0;
}

void hit_player(int how,object who) {
object spot,corpse;

  werror("hit_player("+how+") called in player "+name+" by "
       + who->get_short()+"\n");
  /* Ghosts can't lose hitpoints */
  if(ghost) return;
  current_hp=current_hp-how;
  if(current_hp<0) {
    /* Player died */
    ghost=1;
    /* Lose experience */
    xp=(xp*2)/3;
    remove_room_trigger();
    current_hp=0;
    show_text("You died.");
    if(!frog) {
      oldmodel=model;
      oldtexture=texture;
    }
    else {
      frog=0;
    }
    model="Tghost.ani";
    texture="window_.tm2";
    /* Make sure the new information is saved */
    save();
    game->shout("014 UPDATE "+name+" MODEL "+model+" TEXTURE "+texture+"\n"
              + "003 LOCATE "+name+" X "+x+" Y "+(y-5)+" Z "+z+"\n"
              + "003 LOCATE "+name+" X "+x+" Y "+y+" Z "+z+"\n"
              + "008 "+name+" died!\n");
    /* If possible add a body to the playing field as well */
    drop_dead();
  }
  add_message("012 HP "+current_hp+" MAX "+max_hp+"\n");
}

void set_frog(int how) {
  werror("set_frog("+how+") called in player "+name+"\n");
  if(how==frog) return;
  if(ghost) return;
  frog=how;
  if(frog) {
    oldmodel=model;
    oldtexture=texture;
    model="pfrog.ani";
    texture="items8_.tm2";
  }
  else {
    model=oldmodel;
    texture=oldtexture;
  }  
  game->say(this_object(),
            "014 UPDATE "+name+" MODEL "+model+" TEXTURE "+texture+"\n"
          + "003 LOCATE "+name+" X "+x+" Y "+y+" Z "+z+"\n");
}

void show_text(string what) {
  add_message("008 "+what+"\n");
}
 
string resurrect() {
  werror("resurrect() called in player "+name+"\n");
  model=oldmodel;
  texture=oldtexture;
  ghost=0;
  current_hp=5;
  game->say(this_object(),
              "014 UPDATE "+name+" MODEL "+model+" TEXTURE "+texture+"\n"
            + "003 LOCATE "+name+" X "+x+" Y "+y+" Z "+z+"\n");
  add_message("012 HP "+current_hp+" MAX "+max_hp+"\n");
  return "008 You become alive again.\n";
}

void set_connection(object how) {
int same;

  werror("set_connection() called in "+name+"\n");
  if(!how) {
    /* This means we got disconnected */
    connection=0;
    game->shout("008 "+name+" disconnected from the game!\n");
    return;
  }
  same=0;
  if(connection==how) same=1;
  /* This means we are either a new player or lost our connection */
  connection=how;
  /* Clear the room track so that we look around more */
  roomtrack=({});
  /* And put the current room back into track */
  if(room) set_room(room);
  /* And tell the game we're back */
  if(!same) game->shout("008 "+name+" entered the game!\n");
}

string inventory_list() {
string result;
int i;

  result="016 SIZE "+sizeof(items)+"\n";
  for(i=0;i<sizeof(items);i++) {
    result+="016 INV "+items[i]->get_long()+"\n";
  }
  return result;
}

string score_list() {
string result;

  result="016 SIZE 6\n";
  result+="016 COR gold: "+gold+"\n";
  result+="016 COR hit points: "+current_hp+"/"+max_hp+"\n";
  result+="016 COR spell points: "+current_sp+"/"+max_sp+"\n";
  result+="016 COR attack: "+att+"\n";
  result+="016 COR defense: "+def+"\n";
  result+="016 COR level: "+level+"\n";
  result+="016 COR experience: "+xp+"\n";
  return result;
}

void set_special_item(object how) {
  if(!how) {
    specitem=0;
    game->shout("014 UPDATE "+name+"4 MODEL whand.txt TEXTURE "+texture+"\n");
    return;
  }
  specitem=how;
  game->shout("014 UPDATE "+name+"4 MODEL "+specitem->get_use_model()
            + " TEXTURE "+specitem->get_use_texture()+"\n");
}

string get_player_description() {
string result;

  result="001 NAME "+name+" MODEL "+model+" TEXTURE "+texture+"\n";
  result+="003 LOCATE "+name+" X "+x+" Y "+y+" Z "+z+"\n";
  if(weapon!=this_object()) {
    result+="018 CUPDATE "+name+"2 MODEL "+weapon->get_use_model()+" TEXTURE ";
    result+=weapon->get_use_texture()+"\n";
  }
  if(armour!=this_object()) {
    result+="018 CUPDATE "+name+"3 MODEL "+armour->get_use_model(gender);
    result+=" TEXTURE "+armour->get_use_texture()+"\n";
  }
  if(specitem) {
    result+="018 CUPDATE "+name+"4 MODEL "+specitem->get_use_model();
    result+=" TEXTURE "+specitem->get_use_texture()+"\n";
  }
  return result;
}

string handle_shop(string action,string what) {
object item;
int i;
string ret;

  if(action=="sell") {
    item=present(what);
    if(!item) return "008 Item not in invertory!\n";
    if(item->cant_drop()) return "008 You can't sell this item!\n";
    if(item->get_type()=="torch") return "008 You can't sell a burning item!\n";
    ret=room->handle_shop(this_object(),"sell",item);
    if(ret) {
      remove_item_from_list(item);
      return ret;
    }
    else {
      return "008 Couldn't sell item!\n";
    }
  }
  return room->handle_shop(this_object(),action,what);
}

string get_attack_sound() {
  return weapon->get_sound();
}

int get_attacking() {
  if(weapon->get_defense_modifier()>10) return 1;
  return ::get_attacking();
}

string get_description() {
  if(ghost) return "The ghost of "+name+".\n";
  if(frog) return "A frog looking a bit like "+name+".\n";
  return ::get_description();
}
